# code in this file is adpated from
# https://github.com/ildoonet/pytorch-randaugment/blob/master/RandAugment/augmentations.py
# https://github.com/google-research/fixmatch/blob/master/third_party/auto_augment/augmentations.py
# https://github.com/google-research/fixmatch/blob/master/libml/ctaugment.py
import logging
import random

import numpy as np
import PIL
import PIL.ImageOps
import PIL.ImageEnhance
import PIL.ImageDraw
from PIL import Image
from audioop import bias

logger = logging.getLogger(__name__)

PARAMETER_MAX = 10
RESAMPLE_MODE = None


def AutoContrast(img, **kwarg):
    return PIL.ImageOps.autocontrast(img)


def Brightness(img, v, max_v, bias=0):
    v = _float_parameter(v, max_v) + bias
    return PIL.ImageEnhance.Brightness(img).enhance(v)


def Color(img, v, max_v, bias=0):
    v = _float_parameter(v, max_v) + bias
    return PIL.ImageEnhance.Color(img).enhance(v)


def Contrast(img, v, max_v, bias=0):
    v = _float_parameter(v, max_v) + bias
    return PIL.ImageEnhance.Contrast(img).enhance(v)


def Cutout(img, v, max_v, **kwarg):
    if v == 0:
        return img
    v = _float_parameter(v, max_v)
    v = int(v * min(img.size))
    w, h = img.size
    x0 = np.random.uniform(0, w)
    y0 = np.random.uniform(0, h)
    x0 = int(max(0, x0 - v / 2.))
    y0 = int(max(0, y0 - v / 2.))
    x1 = int(min(w, x0 + v))
    y1 = int(min(h, y0 + v))
    xy = (x0, y0, x1, y1)
    # gray
    color = (127, 127, 127)
    img = img.copy()
    PIL.ImageDraw.Draw(img).rectangle(xy, color)
    return img


def CutoutConst(img, v, max_v, **kwarg):
    v = _int_parameter(v, max_v)
    w, h = img.size
    x0 = np.random.uniform(0, w)
    y0 = np.random.uniform(0, h)
    x0 = int(max(0, x0 - v / 2.))
    y0 = int(max(0, y0 - v / 2.))
    x1 = int(min(w, x0 + v))
    y1 = int(min(h, y0 + v))
    xy = (x0, y0, x1, y1)
    # gray
    color = (127, 127, 127)
    img = img.copy()
    PIL.ImageDraw.Draw(img).rectangle(xy, color)
    return img


def Equalize(img, **kwarg):
    return PIL.ImageOps.equalize(img)


def Identity(img, **kwarg):
    return img


def Invert(img, **kwarg):
    return PIL.ImageOps.invert(img)


def Posterize(img, v, max_v, bias, **kwarg):
    v = _int_parameter(v, max_v) + bias
    return PIL.ImageOps.posterize(img, v)


def Rotate(img, v, max_v, **kwarg):
    v = _float_parameter(v, max_v)
    if random.random() < 0.5:
        v = -v
    return img.rotate(v)


def Sharpness(img, v, max_v, bias):
    v = _float_parameter(v, max_v) + bias
    return PIL.ImageEnhance.Sharpness(img).enhance(v)


def ShearX(img, v, max_v, **kwarg):
    v = _float_parameter(v, max_v)
    if random.random() < 0.5:
        v = -v
    return img.transform(img.size, PIL.Image.AFFINE, (1, v, 0, 0, 1, 0), RESAMPLE_MODE)


def ShearY(img, v, max_v, **kwarg):
    v = _float_parameter(v, max_v)
    if random.random() < 0.5:
        v = -v
    return img.transform(img.size, PIL.Image.AFFINE, (1, 0, 0, v, 1, 0), RESAMPLE_MODE)


def Solarize(img, v, max_v, **kwarg):
    v = _int_parameter(v, max_v)
    return PIL.ImageOps.solarize(img, 256 - v)


def SolarizeAdd(img, v, max_v, threshold=128, **kwarg):
    v = _int_parameter(v, max_v)
    if random.random() < 0.5:
        v = -v
    img_np = np.array(img).astype(np.int)
    img_np = img_np + v
    img_np = np.clip(img_np, 0, 255)
    img_np = img_np.astype(np.uint8)
    img = Image.fromarray(img_np)
    return PIL.ImageOps.solarize(img, threshold)


def TranslateX(img, v, max_v, **kwarg):
    v = _float_parameter(v, max_v)
    if random.random() < 0.5:
        v = -v
    v = int(v * img.size[0])
    return img.transform(img.size, PIL.Image.AFFINE, (1, 0, v, 0, 1, 0), RESAMPLE_MODE)


def TranslateY(img, v, max_v, **kwarg):
    v = _float_parameter(v, max_v)
    if random.random() < 0.5:
        v = -v
    v = int(v * img.size[1])
    return img.transform(img.size, PIL.Image.AFFINE, (1, 0, 0, 0, 1, v), RESAMPLE_MODE)


def TranslateXConst(img, v, max_v, **kwarg):
    v = _float_parameter(v, max_v)
    if random.random() > 0.5:
        v = -v
    return img.transform(img.size, PIL.Image.AFFINE, (1, 0, v, 0, 1, 0), RESAMPLE_MODE)


def TranslateYConst(img, v, max_v, **kwarg):
    v = _float_parameter(v, max_v)
    if random.random() > 0.5:
        v = -v
    return img.transform(img.size, PIL.Image.AFFINE, (1, 0, 0, 0, 1, v), RESAMPLE_MODE)


def _float_parameter(v, max_v):
    return float(v) * max_v / PARAMETER_MAX


def _int_parameter(v, max_v):
    return int(v * max_v / PARAMETER_MAX)


def rand_augment_pool():
    augs = [(AutoContrast, None, None),
            (Brightness, 1.8, 0.1),
            (Color, 1.8, 0.1),
            (Contrast, 1.8, 0.1),
            (CutoutConst, 40, None),
            (Equalize, None, None),
            (Invert, None, None),
            (Posterize, 4, 0),
            (Rotate, 30, None),
            (Sharpness, 1.8, 0.1),
            (ShearX, 0.3, None),
            (ShearY, 0.3, None),
            (Solarize, 256, None),
            (TranslateXConst, 100, None),
            (TranslateYConst, 100, None),
            ]
    return augs


class RandAugment(object):
    def __init__(self, n, m, resample_mode=PIL.Image.BILINEAR):
        assert n >= 1
        assert m >= 1
        global RESAMPLE_MODE
        RESAMPLE_MODE = resample_mode
        self.n = n
        self.m = m
        self.augment_pool = rand_augment_pool()

    def __call__(self, img):
        ops = random.choices(self.augment_pool, k=self.n)
        for op, max_v, bias in ops:
            prob = np.random.uniform(0.2, 0.8)
            if random.random() + prob >= 1:
                img = op(img, v=self.m, max_v=max_v, bias=bias)
        return img

class CutOutAug(object):
    def __init__(self, n, size, resample_mode=PIL.Image.BILINEAR):
        assert n >= 1
        global RESAMPLE_MODE
        RESAMPLE_MODE = resample_mode
        self.n = n
        self.size =size
        self.augment_pool = [(CutoutConst, 40, None)]

    def __call__(self, img):
        ops = random.choices(self.augment_pool, k=self.n)
        for op, max_v, bias in ops:
            prob = np.random.uniform(0.2, 0.8)
            if random.random() + prob >= 1:
                img = op(img, v=self.size, max_v=self.size, bias=bias)
        return img



import argparse
parser = argparse.ArgumentParser()
parser.add_argument('-j', '--workers', default=4, type=int, metavar='N',
                    help='number of data loading workers (default: 4)')
parser.add_argument('--epochs', default=1000, type=int, metavar='N',
                    help='number of total epochs to run')
parser.add_argument('--start-epoch', default=0, type=int, metavar='N',
                    help='manual epoch number (useful on restarts)')
parser.add_argument('-b', '--batch-size', default=256, type=int,
                    metavar='N', help='mini-batch size (default: 128)')
parser.add_argument('--lr', '--learning-rate', default=0.01, type=float,
                    metavar='LR', help='initial learning rate')
parser.add_argument('--momentum', default=0.9, type=float, metavar='M',
                    help='momentum')
parser.add_argument('--weight-decay', '--wd', default=5e-4, type=float,
                    metavar='W', help='weight decay (default: 5e-4)')
parser.add_argument('--print-freq', '-p', default=100, type=int,
                    metavar='N', help='print frequency')
parser.add_argument('--resume', default='', type=str, metavar='PATH',
                    help='path to latest checkpoint (default: none)')
parser.add_argument('-e', '--evaluate', dest='evaluate', action='store_true',
                    help='evaluate model on validation set')
parser.add_argument('--pretrained', dest='pretrained', action='store_true',
                    help='use pre-trained model')
parser.add_argument('--nesterov', action='store_true', help='use nesterov')
parser.add_argument('--save-dir', dest='save_dir',
                    help='The directory used to save the trained models',
                    default='save_temp', type=str)
parser.add_argument('--save-every', dest='save_every',
                    help='Saves checkpoints at every specified number of epochs',
                    type=int, default=10)
parser.add_argument('--dataset', dest='dataset',
                    help='Name of the dataset',
                    default='CUSTOM', type=str)
parser.add_argument('--mode', dest='mode',
                    help='image/text',
                    default='image', type=str)
parser.add_argument('--number-of-class', dest='number_of_class',
                    help='number of class',
                    default=2, type=int)
parser.add_argument('--device', default='gpu', type=str, help='device')
parser.add_argument('--label-smoothing', default=0, type=float, help='label smoothing alpha')
parser.add_argument('--warmup-steps', default=10, type=int, help='warmup steps')
parser.add_argument("--amp", default=True, action="store_true", help="use 16-bit (mixed) precision")
parser.add_argument('--temperature', default=1.25, type=float, help='pseudo label temperature')
parser.add_argument('--threshold', default=0.6, type=float, help='pseudo label threshold')
parser.add_argument('--lambda-u', default=8, type=float, help='coefficient of unlabeled loss')
parser.add_argument('--uda-steps', default=500, type=float, help='warmup steps of lambda-u')
parser.add_argument('--grad-clip', default=0., type=float, help='gradient norm clipping')
parser.add_argument('--resize', default=32, type=int, help='resize image')
parser.add_argument('--data-path', default='./data', type=str, help='data path')
parser.add_argument('--num-labeled', type=int, default=4000, help='number of labeled data')
parser.add_argument('--num-classes', default=10, type=int, help='number of classes')
parser.add_argument("--expand-labels", action="store_true", help="expand labels to fit eval steps")
parser.add_argument("--randaug", nargs="+", type=int, help="use it like this. --randaug 2 10")
parser.add_argument('--dense-dropout', default=0.5, type=float, help='dropout on last dense layer')
parser.add_argument('--dropout', default=0, type=float, help='dropout on layer')
parser.add_argument('--mu', default=7, type=int, help='coefficient of unlabeled batch size')
parser.add_argument('--eval-step', default=1000, type=int, help='number of eval steps to run')
parser.add_argument('--ema', default=0, type=float, help='EMA decay rate')


from copy import deepcopy
import math

import torch
import torch.nn as nn
import torch.nn.functional as F


class ModelEMA(nn.Module):
    def __init__(self, model, decay=0.9999, device=None):
        super().__init__()
        self.module = deepcopy(model)
        self.module.eval()
        self.decay = decay
        self.device = device
        if self.device is not None:
            self.module.to(device=device)

    def forward(self, input):
        return self.module(input)

    def _update(self, model, update_fn):
        with torch.no_grad():
            for ema_v, model_v in zip(self.module.parameters(), model.parameters()):
                if self.device is not None:
                    model_v = model_v.to(device=self.device)
                ema_v.copy_(update_fn(ema_v, model_v))
            for ema_v, model_v in zip(self.module.buffers(), model.buffers()):
                if self.device is not None:
                    model_v = model_v.to(device=self.device)
                ema_v.copy_(model_v)

    def update_parameters(self, model):
        self._update(model, update_fn=lambda e, m: self.decay * e + (1. - self.decay) * m)

    def state_dict(self):
        return self.module.state_dict()

    def load_state_dict(self, state_dict):
        self.module.load_state_dict(state_dict)

class BasicBlock(nn.Module):
    def __init__(self, in_planes, out_planes, stride, dropout=0.0, activate_before_residual=False):
        super(BasicBlock, self).__init__()
        self.bn1 = nn.BatchNorm2d(in_planes, momentum=0.001)
        self.relu1 = nn.LeakyReLU(negative_slope=0.1, inplace=True)
        self.conv1 = nn.Conv2d(in_planes, out_planes, kernel_size=3, stride=stride,
                               padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(out_planes, momentum=0.001)
        self.relu2 = nn.LeakyReLU(negative_slope=0.1, inplace=True)
        self.conv2 = nn.Conv2d(out_planes, out_planes, kernel_size=3, stride=1,
                               padding=1, bias=False)
        self.dropout = dropout
        self.equalInOut = (in_planes == out_planes)
        self.convShortcut = (not self.equalInOut) and nn.Conv2d(in_planes, out_planes,
                                                                kernel_size=1, stride=stride,
                                                                padding=0, bias=False) or None
        self.activate_before_residual = activate_before_residual

    def forward(self, x):
        if not self.equalInOut and self.activate_before_residual == True:
            x = self.relu1(self.bn1(x))
        else:
            out = self.relu1(self.bn1(x))
        out = self.relu2(self.bn2(self.conv1(out if self.equalInOut else x)))
        if self.dropout > 0:
            out = F.dropout(out, p=self.dropout, training=self.training)
        out = self.conv2(out)
        return torch.add(x if self.equalInOut else self.convShortcut(x), out)

class NetworkBlock(nn.Module):
    def __init__(self, nb_layers, in_planes, out_planes, block, stride, dropout=0.0,
                 activate_before_residual=False):
        super(NetworkBlock, self).__init__()
        self.layer = self._make_layer(
            block, in_planes, out_planes, nb_layers, stride, dropout, activate_before_residual)

    def _make_layer(self, block, in_planes, out_planes, nb_layers, stride, dropout,
                    activate_before_residual):
        layers = []
        for i in range(int(nb_layers)):
            layers.append(block(i == 0 and in_planes or out_planes, out_planes,
                                i == 0 and stride or 1, dropout, activate_before_residual))
        return nn.Sequential(*layers)

    def forward(self, x):
        return self.layer(x)

class WideResNet(nn.Module):
    def __init__(self, num_classes, depth=28, widen_factor=2, dropout=0.0, dense_dropout=0.0):
        super(WideResNet, self).__init__()
        channels = [16, 16*widen_factor, 32*widen_factor, 64*widen_factor]
        assert((depth - 4) % 6 == 0)
        n = (depth - 4) / 6
        block = BasicBlock
        # 1st conv before any network block
        self.conv1 = nn.Conv2d(3, channels[0], kernel_size=3, stride=1,
                               padding=1, bias=False)
        # 1st block
        self.block1 = NetworkBlock(
            n, channels[0], channels[1], block, 1, dropout, activate_before_residual=True)
        # 2nd block
        self.block2 = NetworkBlock(
            n, channels[1], channels[2], block, 2, dropout)
        # 3rd block
        self.block3 = NetworkBlock(
            n, channels[2], channels[3], block, 2, dropout)
        # global average pooling and classifier
        self.bn1 = nn.BatchNorm2d(channels[3], momentum=0.001)
        self.relu = nn.LeakyReLU(negative_slope=0.1, inplace=True)
        self.drop = nn.Dropout(dense_dropout)
        self.fc = nn.Linear(channels[3], num_classes)
        self.channels = channels[3]
        print('dropouts are ', dropout, dense_dropout)
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight,
                                        mode='fan_out',
                                        nonlinearity='leaky_relu')
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.constant_(m.weight, 1.0)
                nn.init.constant_(m.bias, 0.0)
            elif isinstance(m, nn.Linear):
                nn.init.xavier_normal_(m.weight)
                nn.init.constant_(m.bias, 0.0)

    def forward(self, x):
        out = self.conv1(x)
        out = self.block1(out)
        out = self.block2(out)
        out = self.block3(out)
        out = self.relu(self.bn1(out))
        out = F.adaptive_avg_pool2d(out, 1)
        out = out.view(-1, self.channels)
        return self.fc(self.drop(out))


def build_wideresnet(args):
    if args.dataset == "CIFAR10":
        depth, widen_factor = 28, 2
    elif args.dataset == 'CIFAR100':
        depth, widen_factor = 28, 8
    else :
        depth, widen_factor = 28, 2

    model = WideResNet(num_classes=args.num_classes,
                       depth=depth,
                       widen_factor=widen_factor,
                       dropout=args.dropout,
                       dense_dropout=args.dense_dropout)

    print(f"Model: WideResNet {depth}x{widen_factor}")
    print(f"Total params: {sum(p.numel() for p in model.parameters())/1e6:.2f}M")
    return model

class SimpleNet(nn.Module):

    def __init__(self):
        super(SimpleNet, self).__init__()
        self.fc1 = nn.Linear(2, 120).double() 
        self.fc2 = nn.Linear(120, 84).double()
        self.fc3 = nn.Linear(84, 2).double()

    def forward(self, x):
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = self.fc3(x)
        return x

def build_simplenet(args):
    model = SimpleNet()
    return model

def get_model(args):
    if args.dataset == "TWOMOON":
        model = build_simplenet(args)
    elif args.dataset == "CIFAR10":
        model = build_wideresnet(args)
    return model

def initialize_model(args):
    if args.model == "wideResNet-28-2":
        depth, widen_factor = 28, 2
        model = WideResNet(num_classes=args.classes,
                       depth=depth,
                       widen_factor=widen_factor,
                       dropout=args.dropout,
                       dense_dropout=args.dense_dropout)

        print(f"Model: WideResNet {depth}-{widen_factor}")
        # print(f"Total params: {sum(p.numel() for p in model.parameters())/1e6:.2f}M")
        return model
    else:
        print("This model is not support yet.")


import logging
import math
import time
import warnings
from itertools import cycle, islice

import torch
import numpy as np
from PIL import Image
from torch.utils.data import Dataset
from torchvision import datasets
from torchvision import transforms
from sklearn.preprocessing import StandardScaler
from sklearn.datasets import make_moons

from .augmentation import RandAugment, CutOutAug
from .autoaugment import CIFAR10Policy, SVHNPolicy
from .utils import AttrDict


cifar10_mean = (0.4914, 0.4822, 0.4465)
cifar10_std = (0.2471, 0.2435, 0.2616)
cifar100_mean = (0.5071, 0.4867, 0.4408)
cifar100_std = (0.2675, 0.2565, 0.2761)
normal_mean = (0.5, 0.5, 0.5)
normal_std = (0.5, 0.5, 0.5)

class MyDataset(Dataset):
    def __init__(self, data, targets, transform=None):
        self.data = data
        self.targets = torch.LongTensor(targets)
        self.transform = transform
        
    def __getitem__(self, index):
        x = self.data[index]
        y = self.targets[index]
        
        if self.transform:
            x = self.transform(x)
        return x, y
    
    def __len__(self):
        return len(self.data)

class TransformToy(object):
    def __init__(self):
        self.ori = transforms.Compose([])
        self.aug = transforms.Compose([])

    def __call__(self, x):
        ori = self.ori(x)
        aug = self.aug(x)
        return ori, aug


def prepare_cifar10(num_labeled=4000, classes=10, eval_step=1000, batch_size=64, resize=32, data_path="./data",randaug=[2,16]):
    args = {"num_labeled": num_labeled,
        "num_classes": classes,
        "eval_step": eval_step, 
        "batch_size": batch_size,
        "resize": resize,
        "data_path": data_path,
        "expand_labels": False,
        "randaug": randaug
       }
    args = AttrDict(args)
    return get_cifar10(args)

def get_toy(args):
    n_samples = 1500
    np.random.seed(0)
    noisy_moons = make_moons(n_samples=n_samples, noise=0.05)
    X, y = noisy_moons
    
    X = StandardScaler().fit_transform(X)
    random_choice = np.random.choice(X.shape[0], 5, replace=False)
    labeled_data = X[random_choice]
    targets  = y[random_choice]

    labeled_dataset = MyDataset(labeled_data, torch.from_numpy(targets).long())
    unlabeled_dataset =MyDataset(X, torch.from_numpy(y).long(), transform=TransformToy())
    test_dataset =MyDataset(X, torch.from_numpy(y).long())

    return labeled_dataset, unlabeled_dataset, test_dataset

def get_cifar10(args):
    transform_labeled = transforms.Compose([
        transforms.RandomHorizontalFlip(),
        CIFAR10Policy(),
        transforms.RandomCrop(size=args.resize,
                              padding=int(args.resize*0.125),
                              padding_mode='reflect'),
        transforms.ToTensor(),
        transforms.Normalize(mean=cifar10_mean, std=cifar10_std)
    ])
    transform_val = transforms.Compose([
        transforms.ToTensor(),
        transforms.Normalize(mean=cifar10_mean, std=cifar10_std)
    ])
    base_dataset = datasets.CIFAR10(args.data_path, train=True, download=True)

    train_labeled_idxs, train_unlabeled_idxs = x_u_split(args, base_dataset.targets)

    train_labeled_dataset = CIFAR10SSL(
        args.data_path, train_labeled_idxs, train=True,
        transform=transform_labeled
    )

    train_unlabeled_dataset = CIFAR10SSL(
        args.data_path, train_unlabeled_idxs,
        train=True,
        transform=TransformMPL(args, mean=cifar10_mean, std=cifar10_std)
    )

    test_dataset = datasets.CIFAR10(args.data_path, train=False,
                                    transform=transform_val, download=False)

    return train_labeled_dataset, train_unlabeled_dataset, test_dataset


def get_cifar100(args):

    transform_labeled = transforms.Compose([
        transforms.RandomHorizontalFlip(),
        CIFAR10Policy(),
        transforms.RandomCrop(size=args.resize,
                              padding=int(args.resize*0.125),
                              padding_mode='reflect'),
        transforms.ToTensor(),
        transforms.Normalize(mean=cifar100_mean, std=cifar100_std)])

    transform_val = transforms.Compose([
        transforms.ToTensor(),
        transforms.Normalize(mean=cifar100_mean, std=cifar100_std)])

    base_dataset = datasets.CIFAR100(args.data_path, train=True, download=True)

    train_labeled_idxs, train_unlabeled_idxs = x_u_split(args, base_dataset.targets)

    train_labeled_dataset = CIFAR100SSL(
        args.data_path, train_labeled_idxs, train=True,
        transform=transform_labeled
    )

    train_unlabeled_dataset = CIFAR100SSL(
        args.data_path, train_unlabeled_idxs, train=True,
        transform=TransformMPL(args, mean=cifar100_mean, std=cifar100_std)
    )

    test_dataset = datasets.CIFAR100(args.data_path, train=False,
                                     transform=transform_val, download=False)

    return train_labeled_dataset, train_unlabeled_dataset, test_dataset


def x_u_split(args, labels):
    label_per_class = args.num_labeled // args.num_classes
    labels = np.array(labels)
    labeled_idx = []
    # unlabeled data: all training data
    unlabeled_idx = np.array(range(len(labels)))
    for i in range(args.num_classes):
        idx = np.where(labels == i)[0]
        idx = np.random.choice(idx, label_per_class, False)
        labeled_idx.extend(idx)
    labeled_idx = np.array(labeled_idx)
    assert len(labeled_idx) == args.num_labeled

    if args.expand_labels or args.num_labeled < args.batch_size:
        num_expand_x = math.ceil(
            args.batch_size * args.eval_step / args.num_labeled)
        labeled_idx = np.hstack([labeled_idx for _ in range(num_expand_x)])
    np.random.shuffle(labeled_idx)
    return labeled_idx, unlabeled_idx


class TransformMPL(object):
    def __init__(self, args, mean, std):
        if args.randaug:
            n, m = args.randaug
        else:
            n, m = 2, 16  # default

        self.ori = transforms.Compose([
            transforms.RandomHorizontalFlip(),
            transforms.RandomCrop(size=args.resize,
                                  padding=int(args.resize*0.125),
                                  padding_mode='reflect')])
        self.aug = transforms.Compose([
            CIFAR10Policy(),
            transforms.RandomHorizontalFlip(),
            transforms.RandomCrop(size=args.resize,
                                  padding=int(args.resize*0.125),
                                  padding_mode='reflect'),
            RandAugment(n=n, m=m),
            CutOutAug(n=1, size=args.resize//4),
            transforms.RandomHorizontalFlip(),
            transforms.RandomCrop(size=args.resize,
                                  padding=int(args.resize*0.125),
                                  padding_mode='reflect')
            ])
        self.normalize = transforms.Compose([
            transforms.ToTensor(),
            transforms.Normalize(mean=mean, std=std)])

    def __call__(self, x):
        ori = self.ori(x)
        aug = self.aug(x)
        return self.normalize(ori), self.normalize(aug)

class CIFAR10SSL(datasets.CIFAR10):
    def __init__(self, root, indexs, train=True,
                 transform=None, target_transform=None,
                 download=False):
        super().__init__(root, train=train,
                         transform=transform,
                         target_transform=target_transform,
                         download=download)
        if indexs is not None:
            self.data = self.data[indexs]
            self.targets = np.array(self.targets)[indexs]

    def __getitem__(self, index):
        img, target = self.data[index], self.targets[index]
        img = Image.fromarray(img)

        if self.transform is not None:
            img = self.transform(img)

        if self.target_transform is not None:
            target = self.target_transform(target)

        return img, target


class CIFAR100SSL(datasets.CIFAR100):
    def __init__(self, root, indexs, train=True,
                 transform=None, target_transform=None,
                 download=False):
        super().__init__(root, train=train,
                         transform=transform,
                         target_transform=target_transform,
                         download=download)
        if indexs is not None:
            self.data = self.data[indexs]
            self.targets = np.array(self.targets)[indexs]

    def __getitem__(self, index):
        img, target = self.data[index], self.targets[index]
        img = Image.fromarray(img)

        if self.transform is not None:
            img = self.transform(img)

        if self.target_transform is not None:
            target = self.target_transform(target)

        return img, target



import os
import shutil
from collections import OrderedDict
import math

import torch
from torch import distributed as dist
from torch import nn
from torch.nn import functional as F
from torch.optim.lr_scheduler import LambdaLR


class AttrDict(dict):
    def __init__(self, *args, **kwargs):
        super(AttrDict, self).__init__(*args, **kwargs)
        self.__dict__ = self

class AverageMeter(object):
    """Computes and stores the average and current value"""
    def __init__(self):
        self.reset()

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count

def accuracy(output, target, topk=(1,)):
    """Computes the precision@k for the specified values of k"""
    maxk = max(topk)
    batch_size = target.size(0)

    _, pred = output.topk(maxk, 1, True, True)
    pred = pred.t()
    correct = pred.eq(target.view(1, -1).expand_as(pred))

    res = []
    for k in topk:
        correct_k = correct[:k].view(-1).float().sum(0)
        res.append(correct_k.mul_(100.0 / batch_size))
    return res

def reduce_tensor(tensor, n):
    rt = tensor.clone()
    dist.all_reduce(rt, op=dist.ReduceOp.SUM)
    rt /= n
    return rt


def get_loss_function(args):
    if args.label_smoothing > 0:
        criterion = SmoothCrossEntropy(alpha=args.label_smoothing)
    else:
        criterion = nn.CrossEntropyLoss()
    return criterion.to(args.device)

def get_cosine_schedule_with_warmup(optimizer,
                                    num_warmup_steps,
                                    num_training_steps,
                                    num_cycles=0.5,
                                    last_epoch=-1):
    def lr_lambda(current_step):
        if current_step < num_warmup_steps:
            return float(current_step) / float(max(1, num_warmup_steps))

        progress = float(current_step - num_warmup_steps) / \
            float(max(1, num_training_steps - num_warmup_steps))
        return max(0.0, 0.5 * (1.0 + math.cos(math.pi * float(num_cycles) * 2.0 * progress)))

    return LambdaLR(optimizer, lr_lambda, last_epoch)

def module_load_state_dict(model, state_dict):
    new_state_dict = OrderedDict()
    for k, v in state_dict.items():
        name = k[7:]  # remove `module.`
        new_state_dict[name] = v
    model.load_state_dict(new_state_dict)


def model_load_state_dict(model, state_dict):
    try:
        model.load_state_dict(state_dict)
    except:
        module_load_state_dict(model, state_dict)


def save_checkpoint(args, state, is_best, finetune=False):
    os.makedirs(args.save_path, exist_ok=True)
    if finetune:
        name = f'{args.name}_finetune'
    else:
        name = args.name
    filename = f'{args.save_path}/{name}_last.pth.tar'
    torch.save(state, filename, _use_new_zipfile_serialization=False)
    if is_best:
        shutil.copyfile(filename, f'{args.save_path}/{args.name}_best.pth.tar')


def accuracy(output, target, topk=(1,)):
    output = output.to(torch.device('cpu'))
    target = target.to(torch.device('cpu'))
    maxk = max(topk)
    batch_size = target.shape[0]

    _, idx = output.sort(dim=1, descending=True)
    pred = idx.narrow(1, 0, maxk).t()
    correct = pred.eq(target.reshape(1, -1).expand_as(pred))

    res = []
    res = []
    for k in topk:
        correct_k = correct[:k].reshape(-1).float().sum(dim=0, keepdim=True)
        res.append(correct_k.mul_(100.0 / batch_size))
    return res


class SmoothCrossEntropy(nn.Module):
    def __init__(self, alpha=0.1):
        super(SmoothCrossEntropy, self).__init__()
        self.alpha = alpha

    def forward(self, logits, labels):
        num_classes = logits.shape[-1]
        alpha_div_k = self.alpha / num_classes
        target_probs = F.one_hot(labels, num_classes=num_classes).float() * \
            (1. - self.alpha) + alpha_div_k
        loss = -(target_probs * torch.log_softmax(logits, dim=-1)).sum(dim=-1)
        return loss.mean()


class AverageMeter(object):
    """Computes and stores the average and current value
       Imported from https://github.com/pytorch/examples/blob/master/imagenet/main.py#L247-L262
    """

    def __init__(self):
        self.reset()

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count



import os
import shutil
import time

import torch
import torch.nn as nn
import torch.nn.parallel
from torch.nn import functional as F
import torch.backends.cudnn as cudnn
import torch.optim
import torch.utils.data
import torchvision.transforms as transforms
import torchvision.datasets as datasets
from torch.cuda import amp

from .model import get_model, ModelEMA
from .data import get_cifar10, get_toy
from .parser_setting import parser
from .utils import get_loss_function, get_cosine_schedule_with_warmup, AverageMeter, accuracy


def initialize_components(args):
    labeled_loader, unlabeled_loader, test_loader = get_dataset_dataloader(args)

    # Check the save_dir exists or not
    if not os.path.exists(args.save_dir):
        os.makedirs(args.save_dir)
    if args.dataset == "CIFAR10" :
        args.number_of_class = 10
    elif args.dataset == 'CIFAR100':
        args.number_of_class = 100
    elif args.dataset == 'TWOMOON':
        args.number_of_class = 2

    # model = torch.nn.DataParallel(get_model(args))
    model = get_model(args)
    if mps_is_available():
        args.device = torch.device("mps")
    elif args.device == "cpu":
        args.device = torch.device("cpu")
    elif args.device == "gpu":
        args.device = torch.device("cuda")
    model.to(args.device)
    # model = ModelEMA(model, 0.995)
    avg_model = None
    if args.ema > 0:
        avg_model = ModelEMA(model, args.ema)
    cudnn.benchmark = True
    criterion = get_loss_function(args)

    first_optimizer = torch.optim.SGD(model.parameters(),
                            lr=args.lr,
                            momentum=args.momentum,
                            weight_decay=args.weight_decay,
                            nesterov=args.nesterov)
    second_optimizer = torch.optim.SGD(model.parameters(),
                            lr=args.lr,
                            momentum=args.momentum,
                            weight_decay=args.weight_decay,
                            nesterov=args.nesterov)
    # first_optimizer = torch.optim.SGD( [{'params': model.parameters(), 'initial_lr': args.lr}] , args.lr,
    #                             momentum=args.momentum,
    #                             weight_decay=args.weight_decay,
    #                             nesterov=args.nesterov)
    # second_optimizer = torch.optim.SGD( [{'params': model.parameters(), 'initial_lr': args.lr}] , args.lr,
    #                             momentum=args.momentum,
    #                             weight_decay=args.weight_decay,
    #                             nesterov=args.nesterov)

    first_scheduler = get_cosine_schedule_with_warmup(first_optimizer,
                                                  0,
                                                  args.epochs)
                                                  

    second_scheduler = get_cosine_schedule_with_warmup(second_optimizer,
                                                  args.warmup_steps,
                                                  args.epochs)

    first_scaler = amp.GradScaler(enabled=args.amp)
    second_scaler = amp.GradScaler(enabled=args.amp)

    return model, avg_model, labeled_loader, unlabeled_loader, test_loader, criterion, first_optimizer, second_optimizer, first_scheduler, second_scheduler, first_scaler, second_scaler

def get_dataset_dataloader(args):
    normalize = transforms.Normalize(mean=[0.485, 0.456, 0.406],
                                         std=[0.229, 0.224, 0.225])
    train_transformation_list = []
    test_transformation_list = []
    if args.mode == 'image':
        train_transformation_list += [
                # transforms.RandomHorizontalFlip(),
                # transforms.RandomCrop(32, 4),
                transforms.Resize((128,128)),
                transforms.ToTensor(),
                normalize]
        test_transformation_list += [
                # transforms.RandomHorizontalFlip(),
                # transforms.RandomCrop(32, 4),
                transforms.Resize((128,128)),
                transforms.ToTensor(),
                normalize]

    train_transformations = transforms.Compose(train_transformation_list)
    test_transformations = transforms.Compose(test_transformation_list)
    if args.dataset == "CIFAR10" :
        labeled_dataset, unlabeled_dataset, test_dataset = get_cifar10(args)

    elif args.dataset == 'CIFAR100':
        train_dataset = datasets.CIFAR100(root='./data', train=True, transform=train_transformations, download=True)
        test_dataset = datasets.CIFAR100(root='./data', train=False, transform=test_transformations, download=True)

    elif args.dataset == 'TWOMOON':
        labeled_dataset, unlabeled_dataset, test_dataset = get_toy(args)

    labeled_loader = torch.utils.data.DataLoader(
            labeled_dataset,
            batch_size=args.batch_size, shuffle=True,
            num_workers=args.workers, pin_memory=True)

    unlabeled_loader = torch.utils.data.DataLoader(
            unlabeled_dataset,
            batch_size=args.batch_size*args.mu, shuffle=True,
            num_workers=args.workers, pin_memory=True)

    test_loader = torch.utils.data.DataLoader(
            test_dataset,
            batch_size=128, shuffle=False,
            num_workers=args.workers, pin_memory=True)

    return labeled_loader, unlabeled_loader, test_loader

def mps_is_available():
    if not torch.backends.mps.is_available():
        return False
    else:
        return True

def train_loop(args, labeled_loader, unlabeled_loader, test_loader, model, avg_model, criterion, first_optimizer, second_optimizer, first_scheduler, second_scheduler, first_scaler, second_scaler, best_top1):
    moving_dot_product = torch.empty(1).to(args.device)
    limit = 3.0**(0.5)  # 3 = 6 / (f_in + f_out)
    nn.init.uniform_(moving_dot_product, -limit, limit)

    model.zero_grad()

    

    labeled_iter = iter(labeled_loader)
    unlabeled_iter = iter(unlabeled_loader)

    batch_time = AverageMeter()
    losses = AverageMeter()
    top1s = AverageMeter()
    
    start = time.time()
    end = time.time()

    for epoch in range(args.start_epoch, args.epochs):
            model.train()
            try:
                images_l, target = next(labeled_iter)
            except:
                labeled_iter = iter(labeled_loader)
                images_l, target = next(labeled_iter)
            try:
                (images_uw, images_us), _ = next(unlabeled_iter)
            except:
                unlabeled_iter = iter(unlabeled_loader)
                (images_uw, images_us), _ = next(unlabeled_iter)

            images_l = images_l.to(args.device)
            images_uw = images_uw.to(args.device)
            images_us = images_us.to(args.device)
            target = target.to(args.device)

            with amp.autocast(enabled=args.amp):
                batch_size = images_l.shape[0]
                t_images = torch.cat((images_l, images_uw, images_us))
                t_logits = model(t_images)
                t_logits_l = t_logits[:batch_size]
                t_logits_uw, t_logits_us = t_logits[batch_size:].chunk(2)

                t_loss_l = criterion(t_logits_l, target)

                soft_pseudo_label = torch.softmax(t_logits_uw.detach()/args.temperature, dim=-1)
                max_probs, hard_pseudo_label = torch.max(soft_pseudo_label, dim=-1)
                mask = max_probs.ge(args.threshold).float()
                t_loss_u = torch.mean(
                    -(soft_pseudo_label * torch.log_softmax(t_logits_us, dim=-1)).sum(dim=-1) * mask
                )
                weight_u = args.lambda_u * min(1., (epoch+1) / args.uda_steps)
                t_loss_uda = t_loss_l + weight_u * t_loss_u
                # t_loss_uda = t_loss_l
                s_logits = t_logits
                s_logits_l = t_logits_l
                s_logits_us = t_logits_us

                del s_logits

                s_loss_l_old = F.cross_entropy(s_logits_l.detach(), target)
                s_loss = criterion(s_logits_us, hard_pseudo_label)

            if args.amp:
                first_scaler.scale(s_loss).backward(retain_graph=True)
                first_scaler.unscale_(first_optimizer)
            else:
                first_optimizer.zero_grad()
                second_optimizer.zero_grad()
                s_loss.requires_grad_(True)
                s_loss.backward(retain_graph=True)

            clone_lr = first_optimizer.param_groups[0]['lr'] 
            clone_grad = {}
            clone_data = {}

            if args.gpu:          
                if not sum (v for v in first_scaler._found_inf_per_device(first_optimizer).values() ):
                    for name, p in model.named_parameters():
                        if p.data is not None and p.grad is not None:
                            # p.grad = torch.clamp(p.grad, min=-.1, max=.1)
                            clone_grad[name] = p.grad.clone().detach()
                            clone_data[name] = p.data.clone()
                            p.data -= clone_lr * clone_grad[name]   
                            p.grad.data.zero_()
                else:
                    print('inf values')
            else:
                for name, p in model.named_parameters():
                    if p.data is not None and p.grad is not None:
                        # p.grad = torch.clamp(p.grad, min=-.1, max=.1)
                        clone_grad[name] = p.grad.clone().detach()
                        clone_data[name] = p.data.clone()
                        p.data -= clone_lr * clone_grad[name]   
                        p.grad.data.zero_()


            if args.amp:
                first_scaler.update()
            first_scheduler.step()

            if args.ema > 0:
                avg_model.update_parameters(model)

            with amp.autocast(enabled=args.amp):
                with torch.no_grad(): 
                    s_logits_l = model(images_l)
                s_loss_l_new = F.cross_entropy(s_logits_l.detach(), target)
                dot_product = s_loss_l_new - s_loss_l_old
                moving_dot_product = moving_dot_product * 0.99 + dot_product * 0.01
                dot_product = dot_product - moving_dot_product
                _, hard_pseudo_label = torch.max(t_logits_us.detach(), dim=-1)
                t_loss_mpl = dot_product * F.cross_entropy(t_logits_us , hard_pseudo_label)
                t_loss = t_loss_uda + 1*t_loss_mpl

            if args.amp:
                second_scaler.scale(t_loss).backward()
            else:
                t_loss.requires_grad_(True)
                t_loss.backward()

            if args.grad_clip > 0:
                second_scaler.unscale_(second_optimizer)
                nn.utils.clip_grad_norm_(model.parameters(), args.grad_clip)

            if args.amp:
                second_scaler.step(second_optimizer)
                second_scaler.update()
            else:
                second_optimizer.step()

            second_scheduler.step()

            model.zero_grad()

            output = s_logits_l.detach().float()
            loss = s_loss_l_new.float()
            # Measure accuracy and record loss
            top1 = accuracy(output.data, target)[0]
            losses.update(loss.item(), target.size(0))
            top1s.update(top1.item(), target.size(0))

            # measure elapsed time
            batch_time.update(time.time() - end)
            end = time.time()

            if epoch % args.print_freq == 0:
                print('Epoch: [{0}]\t'
                      'Elapsed: {batch_time.val:.3f}s ({batch_time.avg:.3f}s)\t'
                      'Loss: {loss.val:.4f} {t_loss: .4f} \t'
                      'Top1 Accuracy: {top1.val:.3f}% '.format(
                          epoch, batch_time=batch_time,
                        loss=losses, top1=top1s, t_loss=t_loss.float().item()))
                # print("   Epoch total time: {total_time:.3f}s".format(total_time=time.time()-start))

            if epoch % 1000 == 0 or epoch == args.epochs:
                # top1 = validate_old(args, test_loader, model, criterion)
                test_model = avg_model if avg_model is not None else model
                top1 = validate(args, test_loader, test_model, criterion)
                is_best = top1 >= best_top1
                best_top1 = max(top1, best_top1)

                print(f"    Best top-1 acc: {best_top1:.2f}%")

                if epoch > 0 and epoch % args.save_every == 0:
                    save_checkpoint({
                        'epoch': epoch + 1,
                        'state_dict': model.state_dict(),
                        'best_top1': best_top1,
                        'first_optimizer': first_optimizer.state_dict(),
                        'second_optimizer': second_optimizer.state_dict(),
                        'avg_state_dict': avg_model.state_dict() if avg_model is not None else None
                    }, is_best, filename=os.path.join(args.save_dir, 'checkpoint.th'))

                if is_best:
                    save_checkpoint({
                        'epoch': epoch + 1,
                        'state_dict': model.state_dict(),
                        'best_top1': best_top1,
                        'first_optimizer': first_optimizer.state_dict(),
                        'second_optimizer': second_optimizer.state_dict(),
                        'avg_state_dict': avg_model.state_dict() if avg_model is not None else None
                    }, is_best, filename=os.path.join(args.save_dir, 'best.th'))

def main():
    args = parser.parse_args()
    best_top1 = 0
    model, avg_model, labeled_loader, unlabeled_loader, test_loader, criterion, first_optimizer, second_optimizer, first_scheduler, second_scheduler, first_scaler, second_scaler = initialize_components(args)
    if args.resume:
        if os.path.isfile(args.resume):
            print("loading checkpoint '{}'".format(args.resume))
            checkpoint = torch.load(args.resume)
            args.start_epoch = checkpoint['epoch']
            best_top1 = checkpoint['best_top1']
            first_optimizer = checkpoint['first_optimizer']
            second_optimizer = checkpoint['second_optimizer']
            model.load_state_dict(checkpoint['state_dict'])
            print("checkpoint loaded'{}' (epoch {})"
                  .format(args.evaluate, checkpoint['epoch']))
            print('checkpoint epoch:', checkpoint['epoch'])
            print('resume lr {:.5e}'.format(optimizer.param_groups[0]['lr']))
            if avg_model is not None:
                    model_load_state_dict(avg_model, checkpoint['avg_state_dict'])
        else:
            print("no checkpoint found at '{}'".format(args.resume)) 
            return

    if args.evaluate:
        validate(test_loader, model, criterion)
        return
        
    train_loop(args, labeled_loader, unlabeled_loader, test_loader, model, avg_model, criterion, first_optimizer, second_optimizer, first_scheduler, second_scheduler, first_scaler, second_scaler, best_top1)
    return


def train(args, labeled_iter, unlabeled_iter, model, avg_model, criterion, first_optimizer, second_optimizer, first_scheduler, second_scheduler, first_scaler, second_scaler, moving_dot_product, epoch):
    batch_time = AverageMeter()
    losses = AverageMeter()
    top1s = AverageMeter()
    model.train()
    start = time.time()
    end = time.time()

    for i, (input, target) in enumerate(train_loader):
        target = target.to(args.device)
        target_var = target

        # compute output
        if args.mode == 'text':
            mask = input['attention_mask'].to(args.device)
            input_id = input['input_ids'].squeeze(1).to(args.device)
            output = model(input_id, mask)
        else:
            input_var = input.to(args.device)

            images_l = images_l.to(args.device)
            images_uw = images_uw.to(args.device)
            images_us = images_us.to(args.device)
            targets = targets.to(args.device)
            output = model(input_var)
        loss = criterion(output, target_var) 
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()
        output = output.float()
        loss = loss.float()
        # measure accuracy and record loss
        top1 = accuracy(output.data, target)[0]
        losses.update(loss.item(), target.size(0))
        top1s.update(top1.item(), target.size(0))

        # measure elapsed time
        batch_time.update(time.time() - end)
        end = time.time()

        if i % args.print_freq == 0:
            print('Epoch: [{0}][{1}/{2}]\t'
                  'Elapsed: {batch_time.val:.3f}s ({batch_time.avg:.3f}s)\t'
                  'Loss: {loss.val:.4f} ({loss.avg:.4f})\t'
                  'Top1 Accuracy: {top1.val:.3f}% ({top1.avg:.3f}%)'.format(
                      epoch, i, len(train_loader), batch_time=batch_time,
                    loss=losses, top1=top1s))
    print("Epoch total time: {total_time:.3f}s".format(total_time=time.time()-start))


def validate(args, test_loader, model, criterion):
    batch_time = AverageMeter()
    losses = AverageMeter()
    top1s = AverageMeter()
    model.eval()

    end = time.time()
    with torch.no_grad():
        for i, (input, target) in enumerate(test_loader):
            target = target.to(args.device)
            input_var = input.to(args.device)
            target_var = target.to(args.device)

            with amp.autocast(enabled=args.amp):
                # compute output
                if args.mode == 'text':
                    mask = input['attention_mask'].to(args.device)
                    input_id = input['input_ids'].squeeze(1).to(args.device)
                    output = model(input_id, mask)
                else:
                    input_var = input.to(args.device)
                    output = model(input_var)
                loss = criterion(output, target_var)
                output = output.float()
                loss = loss.float()

            # measure accuracy and record loss
            top1 = accuracy(output.data, target)[0]
            losses.update(loss.item(), target.size(0))
            top1s.update(top1.item(), target.size(0))

            # measure elapsed time
            batch_time.update(time.time() - end)
            end = time.time()

    print('Validation Top1 Accuracy: {top1.avg:.3f}%'
          .format(top1=top1s))

    return top1s.avg

def validate_old(args, test_loader, model, criterion):
    batch_time = AverageMeter()
    losses = AverageMeter()
    top1s = AverageMeter()
    model.eval()

    end = time.time()
    with torch.no_grad():
        for i, (input, target) in enumerate(test_loader):
            target = target.to(args.device)
            input_var = input.to(args.device)
            target_var = target.to(args.device)

            # compute output
            if args.mode == 'text':
                mask = input['attention_mask'].to(args.device)
                input_id = input['input_ids'].squeeze(1).to(args.device)
                output = model(input_id, mask)
            else:
                input_var = input.to(args.device)
                output = model(input_var)
            loss = criterion(output, target_var)
            output = output.float()
            loss = loss.float()

            # measure accuracy and record loss
            top1 = accuracy(output.data, target)[0]
            losses.update(loss.item(), target.size(0))
            top1s.update(top1.item(), target.size(0))

            # measure elapsed time
            batch_time.update(time.time() - end)
            end = time.time()

    print('Validation Top1 Accuracy: {top1.avg:.3f}%'
          .format(top1=top1s))

    return top1s.avg

def save_checkpoint(state, is_best, filename='checkpoint.pth.tar'):
    torch.save(state, filename)

def predict(args, input, model, transform=None):
    model.eval()
    if transform is None:
        cifar10_mean = (0.4914, 0.4822, 0.4465)
        cifar10_std = (0.2471, 0.2435, 0.2616)
        transform = transforms.Compose([
            transforms.ToTensor(),
            transforms.Normalize(mean=cifar10_mean, std=cifar10_std)
        ])
        input = transform(input)
    with torch.no_grad():
            if input.ndim == 3:
                input = input.unsqueeze(0)
            input_var = input.to(args.device)
            output = model(input_var)
        
    return output.cpu().numpy()



if __name__ == '__main__':
    main()



from PIL import Image, ImageEnhance, ImageOps
import numpy as np
import random


class ImageNetPolicy(object):
    """ Randomly choose one of the best 24 Sub-policies on ImageNet.

        Example:
        >>> policy = ImageNetPolicy()
        >>> transformed = policy(image)

        Example as a PyTorch Transform:
        >>> transform=transforms.Compose([
        >>>     transforms.Resize(256),
        >>>     ImageNetPolicy(),
        >>>     transforms.ToTensor()])
    """
    def __init__(self, fillcolor=(128, 128, 128)):
        self.policies = [
            SubPolicy(0.4, "posterize", 8, 0.6, "rotate", 9, fillcolor),
            SubPolicy(0.6, "solarize", 5, 0.6, "autocontrast", 5, fillcolor),
            SubPolicy(0.8, "equalize", 8, 0.6, "equalize", 3, fillcolor),
            SubPolicy(0.6, "posterize", 7, 0.6, "posterize", 6, fillcolor),
            SubPolicy(0.4, "equalize", 7, 0.2, "solarize", 4, fillcolor),

            SubPolicy(0.4, "equalize", 4, 0.8, "rotate", 8, fillcolor),
            SubPolicy(0.6, "solarize", 3, 0.6, "equalize", 7, fillcolor),
            SubPolicy(0.8, "posterize", 5, 1.0, "equalize", 2, fillcolor),
            SubPolicy(0.2, "rotate", 3, 0.6, "solarize", 8, fillcolor),
            SubPolicy(0.6, "equalize", 8, 0.4, "posterize", 6, fillcolor),

            SubPolicy(0.8, "rotate", 8, 0.4, "color", 0, fillcolor),
            SubPolicy(0.4, "rotate", 9, 0.6, "equalize", 2, fillcolor),
            SubPolicy(0.0, "equalize", 7, 0.8, "equalize", 8, fillcolor),
            SubPolicy(0.6, "invert", 4, 1.0, "equalize", 8, fillcolor),
            SubPolicy(0.6, "color", 4, 1.0, "contrast", 8, fillcolor),

            SubPolicy(0.8, "rotate", 8, 1.0, "color", 2, fillcolor),
            SubPolicy(0.8, "color", 8, 0.8, "solarize", 7, fillcolor),
            SubPolicy(0.4, "sharpness", 7, 0.6, "invert", 8, fillcolor),
            SubPolicy(0.6, "shearX", 5, 1.0, "equalize", 9, fillcolor),
            SubPolicy(0.4, "color", 0, 0.6, "equalize", 3, fillcolor),

            SubPolicy(0.4, "equalize", 7, 0.2, "solarize", 4, fillcolor),
            SubPolicy(0.6, "solarize", 5, 0.6, "autocontrast", 5, fillcolor),
            SubPolicy(0.6, "invert", 4, 1.0, "equalize", 8, fillcolor),
            SubPolicy(0.6, "color", 4, 1.0, "contrast", 8, fillcolor),
            SubPolicy(0.8, "equalize", 8, 0.6, "equalize", 3, fillcolor)
        ]


    def __call__(self, img):
        policy_idx = random.randint(0, len(self.policies) - 1)
        return self.policies[policy_idx](img)

    def __repr__(self):
        return "AutoAugment ImageNet Policy"


class CIFAR10Policy(object):
    """ Randomly choose one of the best 25 Sub-policies on CIFAR10.

        Example:
        >>> policy = CIFAR10Policy()
        >>> transformed = policy(image)

        Example as a PyTorch Transform:
        >>> transform=transforms.Compose([
        >>>     transforms.Resize(256),
        >>>     CIFAR10Policy(),
        >>>     transforms.ToTensor()])
    """
    def __init__(self, fillcolor=(128, 128, 128)):
        self.policies = [
            SubPolicy(0.1, "invert", 7, 0.2, "contrast", 6, fillcolor),
            SubPolicy(0.7, "rotate", 2, 0.3, "translateX", 9, fillcolor),
            SubPolicy(0.8, "sharpness", 1, 0.9, "sharpness", 3, fillcolor),
            SubPolicy(0.5, "shearY", 8, 0.7, "translateY", 9, fillcolor),
            SubPolicy(0.5, "autocontrast", 8, 0.9, "equalize", 2, fillcolor),

            SubPolicy(0.2, "shearY", 7, 0.3, "posterize", 7, fillcolor),
            SubPolicy(0.4, "color", 3, 0.6, "brightness", 7, fillcolor),
            SubPolicy(0.3, "sharpness", 9, 0.7, "brightness", 9, fillcolor),
            SubPolicy(0.6, "equalize", 5, 0.5, "equalize", 1, fillcolor),
            SubPolicy(0.6, "contrast", 7, 0.6, "sharpness", 5, fillcolor),

            SubPolicy(0.7, "color", 7, 0.5, "translateX", 8, fillcolor),
            SubPolicy(0.3, "equalize", 7, 0.4, "autocontrast", 8, fillcolor),
            SubPolicy(0.4, "translateY", 3, 0.2, "sharpness", 6, fillcolor),
            SubPolicy(0.9, "brightness", 6, 0.2, "color", 8, fillcolor),
            SubPolicy(0.5, "solarize", 2, 0.0, "invert", 3, fillcolor),

            SubPolicy(0.2, "equalize", 0, 0.6, "autocontrast", 0, fillcolor),
            SubPolicy(0.2, "equalize", 8, 0.6, "equalize", 4, fillcolor),
            SubPolicy(0.9, "color", 9, 0.6, "equalize", 6, fillcolor),
            SubPolicy(0.8, "autocontrast", 4, 0.2, "solarize", 8, fillcolor),
            SubPolicy(0.1, "brightness", 3, 0.7, "color", 0, fillcolor),

            SubPolicy(0.4, "solarize", 5, 0.9, "autocontrast", 3, fillcolor),
            SubPolicy(0.9, "translateY", 9, 0.7, "translateY", 9, fillcolor),
            SubPolicy(0.9, "autocontrast", 2, 0.8, "solarize", 3, fillcolor),
            SubPolicy(0.8, "equalize", 8, 0.1, "invert", 3, fillcolor),
            SubPolicy(0.7, "translateY", 9, 0.9, "autocontrast", 1, fillcolor)
        ]


    def __call__(self, img):
        policy_idx = random.randint(0, len(self.policies) - 1)
        return self.policies[policy_idx](img)

    def __repr__(self):
        return "AutoAugment CIFAR10 Policy"


class SVHNPolicy(object):
    """ Randomly choose one of the best 25 Sub-policies on SVHN.

        Example:
        >>> policy = SVHNPolicy()
        >>> transformed = policy(image)

        Example as a PyTorch Transform:
        >>> transform=transforms.Compose([
        >>>     transforms.Resize(256),
        >>>     SVHNPolicy(),
        >>>     transforms.ToTensor()])
    """
    def __init__(self, fillcolor=(128, 128, 128)):
        self.policies = [
            SubPolicy(0.9, "shearX", 4, 0.2, "invert", 3, fillcolor),
            SubPolicy(0.9, "shearY", 8, 0.7, "invert", 5, fillcolor),
            SubPolicy(0.6, "equalize", 5, 0.6, "solarize", 6, fillcolor),
            SubPolicy(0.9, "invert", 3, 0.6, "equalize", 3, fillcolor),
            SubPolicy(0.6, "equalize", 1, 0.9, "rotate", 3, fillcolor),

            SubPolicy(0.9, "shearX", 4, 0.8, "autocontrast", 3, fillcolor),
            SubPolicy(0.9, "shearY", 8, 0.4, "invert", 5, fillcolor),
            SubPolicy(0.9, "shearY", 5, 0.2, "solarize", 6, fillcolor),
            SubPolicy(0.9, "invert", 6, 0.8, "autocontrast", 1, fillcolor),
            SubPolicy(0.6, "equalize", 3, 0.9, "rotate", 3, fillcolor),

            SubPolicy(0.9, "shearX", 4, 0.3, "solarize", 3, fillcolor),
            SubPolicy(0.8, "shearY", 8, 0.7, "invert", 4, fillcolor),
            SubPolicy(0.9, "equalize", 5, 0.6, "translateY", 6, fillcolor),
            SubPolicy(0.9, "invert", 4, 0.6, "equalize", 7, fillcolor),
            SubPolicy(0.3, "contrast", 3, 0.8, "rotate", 4, fillcolor),

            SubPolicy(0.8, "invert", 5, 0.0, "translateY", 2, fillcolor),
            SubPolicy(0.7, "shearY", 6, 0.4, "solarize", 8, fillcolor),
            SubPolicy(0.6, "invert", 4, 0.8, "rotate", 4, fillcolor),
            SubPolicy(0.3, "shearY", 7, 0.9, "translateX", 3, fillcolor),
            SubPolicy(0.1, "shearX", 6, 0.6, "invert", 5, fillcolor),

            SubPolicy(0.7, "solarize", 2, 0.6, "translateY", 7, fillcolor),
            SubPolicy(0.8, "shearY", 4, 0.8, "invert", 8, fillcolor),
            SubPolicy(0.7, "shearX", 9, 0.8, "translateY", 3, fillcolor),
            SubPolicy(0.8, "shearY", 5, 0.7, "autocontrast", 3, fillcolor),
            SubPolicy(0.7, "shearX", 2, 0.1, "invert", 5, fillcolor)
        ]


    def __call__(self, img):
        policy_idx = random.randint(0, len(self.policies) - 1)
        return self.policies[policy_idx](img)

    def __repr__(self):
        return "AutoAugment SVHN Policy"


class SubPolicy(object):
    def __init__(self, p1, operation1, magnitude_idx1, p2, operation2, magnitude_idx2, fillcolor=(128, 128, 128)):
        ranges = {
            "shearX": np.linspace(0, 0.3, 10),
            "shearY": np.linspace(0, 0.3, 10),
            "translateX": np.linspace(0, 150 / 331, 10),
            "translateY": np.linspace(0, 150 / 331, 10),
            "rotate": np.linspace(0, 30, 10),
            "color": np.linspace(0.0, 0.9, 10),
            "posterize": np.round(np.linspace(8, 4, 10), 0).astype(np.int),
            "solarize": np.linspace(256, 0, 10),
            "contrast": np.linspace(0.0, 0.9, 10),
            "sharpness": np.linspace(0.0, 0.9, 10),
            "brightness": np.linspace(0.0, 0.9, 10),
            "autocontrast": [0] * 10,
            "equalize": [0] * 10,
            "invert": [0] * 10
        }

        # from https://stackoverflow.com/questions/5252170/specify-image-filling-color-when-rotating-in-python-with-pil-and-setting-expand
        def rotate_with_fill(img, magnitude):
            rot = img.convert("RGBA").rotate(magnitude)
            return Image.composite(rot, Image.new("RGBA", rot.size, (128,) * 4), rot).convert(img.mode)

        func = {
            "shearX": lambda img, magnitude: img.transform(
                img.size, Image.AFFINE, (1, magnitude * random.choice([-1, 1]), 0, 0, 1, 0),
                Image.BICUBIC, fillcolor=fillcolor),
            "shearY": lambda img, magnitude: img.transform(
                img.size, Image.AFFINE, (1, 0, 0, magnitude * random.choice([-1, 1]), 1, 0),
                Image.BICUBIC, fillcolor=fillcolor),
            "translateX": lambda img, magnitude: img.transform(
                img.size, Image.AFFINE, (1, 0, magnitude * img.size[0] * random.choice([-1, 1]), 0, 1, 0),
                fillcolor=fillcolor),
            "translateY": lambda img, magnitude: img.transform(
                img.size, Image.AFFINE, (1, 0, 0, 0, 1, magnitude * img.size[1] * random.choice([-1, 1])),
                fillcolor=fillcolor),
            "rotate": lambda img, magnitude: rotate_with_fill(img, magnitude),
            "color": lambda img, magnitude: ImageEnhance.Color(img).enhance(1 + magnitude * random.choice([-1, 1])),
            "posterize": lambda img, magnitude: ImageOps.posterize(img, magnitude),
            "solarize": lambda img, magnitude: ImageOps.solarize(img, magnitude),
            "contrast": lambda img, magnitude: ImageEnhance.Contrast(img).enhance(
                1 + magnitude * random.choice([-1, 1])),
            "sharpness": lambda img, magnitude: ImageEnhance.Sharpness(img).enhance(
                1 + magnitude * random.choice([-1, 1])),
            "brightness": lambda img, magnitude: ImageEnhance.Brightness(img).enhance(
                1 + magnitude * random.choice([-1, 1])),
            "autocontrast": lambda img, magnitude: ImageOps.autocontrast(img),
            "equalize": lambda img, magnitude: ImageOps.equalize(img),
            "invert": lambda img, magnitude: ImageOps.invert(img)
        }

        self.p1 = p1
        self.operation1 = func[operation1]
        self.magnitude1 = ranges[operation1][magnitude_idx1]
        self.p2 = p2
        self.operation2 = func[operation2]
        self.magnitude2 = ranges[operation2][magnitude_idx2]


    def __call__(self, img):
        if random.random() < self.p1: img = self.operation1(img, self.magnitude1)
        if random.random() < self.p2: img = self.operation2(img, self.magnitude2)
        return img


from typing import Optional
import os

import torch
from torch.cuda import amp

from .model import initialize_model, ModelEMA
from .utils import get_loss_function, get_cosine_schedule_with_warmup
from .main import train_loop, validate, predict
from .data import prepare_cifar10

def mps_is_available():
    if not torch.backends.mps.is_available():
        return False
    else:
        return True

def round_to_multiple(number, multiple):
    return multiple * round(number / multiple)

class MegaBoost:
    def __init__(self, config: Optional[dict] = None):
        self.mode = "train"
        self.epochs = 300000
        self.lr = 0.05
        self.model = "wideResNet-28-2"
        self.classes = 10

        self.device = "gpu"
        self.momentum = 0.9
        self.nesterov = True
        self.label_smoothing = 0.15 
        self.warmup_steps = 50
        self.amp = True
        self.print_freq = round_to_multiple(int(self.epochs*0.000333333333), 10)
        self.temperature = 1.25
        self.threshold = 0.6
        self.dense_dropout = 0.5
        self.dropout = 0
        self.eval_step = round_to_multiple(int(self.epochs*0.00333333333), 10)
        self.ema = 0.995
        self.weight_decay = 5e-4
        self.best_top1 = 0 
        self.save_dir = "result"
        self.start_epoch = 0
        self.lambda_u = 8
        self.uda_steps = round_to_multiple(int(self.epochs * 0.01666666), 10)
        self.grad_clip = 0
        self.save_every = 1
        self.gpu = False

        if config is not None:
            for item in config:
                setattr(self, item, config[item])

        self.model = initialize_model(self)

        if not os.path.exists(self.save_dir):
            os.makedirs(self.save_dir)

        if self.mode == "train" and self.device != "gpu":
            print("Stop: currently semi-supervised mode supports gpu only.")
            return

        if mps_is_available():
            self.device = torch.device("mps")
            self.amp = False
            print("Mps is available")
        elif self.device == "cpu":
            self.device = torch.device("cpu")
            self.amp = False
        elif self.device == "gpu":
            self.device = torch.device("cuda")
            self.gpu = True
            torch.backends.cudnn.benchmark = True


        self.model.to(self.device)

        self.avg_model = None
        if self.ema > 0:
            self.avg_model = ModelEMA(self.model, self.ema)
        self.criterion = get_loss_function(self)

        if self.mode == "train":
            self.first_optimizer = torch.optim.SGD(self.model.parameters(),
                            lr=self.lr,
                            momentum=self.momentum,
                            weight_decay=self.weight_decay,
                            nesterov=self.nesterov)

            self.second_optimizer = torch.optim.SGD(self.model.parameters(),
                            lr=self.lr,
                            momentum=self.momentum,
                            weight_decay=self.weight_decay,
                            nesterov=self.nesterov)
            
            self.first_scheduler = get_cosine_schedule_with_warmup(self.first_optimizer,
                                                  0,
                                                  self.epochs)
                                                  

            self.second_scheduler = get_cosine_schedule_with_warmup(self.second_optimizer,
                                                  self.warmup_steps,
                                                  self.epochs)

            self.first_scaler = amp.GradScaler(enabled=self.amp)
            self.second_scaler = amp.GradScaler(enabled=self.amp)

    def fit(self, X, T, UX = None):
        if self.mode == "train":
            train_loop(self, X, UX, T, self.model, self.avg_model, self.criterion, self.first_optimizer, self.second_optimizer, self.first_scheduler, self.second_scheduler, self.first_scaler, self.second_scaler, self.best_top1)
            return
        return

    def validate(self, X):
        test_model = self.avg_model if self.avg_model is not None else self.model
        validate(self, X, test_model, self.criterion)
        return

    def predict(self, input, transform = None):
        test_model = self.avg_model if self.avg_model is not None else self.model
        res = predict(self, input, test_model, transform=transform)
        return res


from setuptools import setup

setup(
    name='megaboost',
    version='0.1.1',    
    description='Semi-supervised AutoML library on top of Pytorch',
    url='',
    author='Max Ng',
    author_email='maxnghello@gmail.com',
    license='',
    packages=['megaboost'],
    install_requires=['numpy>=1.19.2',
                        'Pillow>=9.2.0',
                        'scikit_learn>=1.2.0',
                        'torch>=1.12.1',
                        'torchvision>=0.13.1'                   
                      ],

    classifiers=[
        'Development Status :: 1 - Planning',
        'Intended Audience :: Science/Research',
        'Operating System :: POSIX :: Linux',        
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.10'
    ],
)